from typing import TYPE_CHECKING

from middlewared.api.base.handler.accept import model_dict_from_list
from middlewared.api.base.handler.dump_params import dump_params
from middlewared.api.base.handler.version import APIVersionsAdapter, APIVersionDoesNotContainModelException
from middlewared.api.base.server.method import Method
from middlewared.utils.service.call import MethodNotFoundError
from middlewared.utils.service.crud import real_crud_method

if TYPE_CHECKING:
    from middlewared.api.base.server.ws_handler.rpc import RpcWebSocketApp
    from middlewared.main import Middleware


class LegacyAPIMethod(Method):
    """
    Represents a middleware legacy API method used in JSON-RPC server. Converts method parameters and return value
    between most recent API version (used in the code) and predetermined legacy API version.
    """

    def __init__(self, middleware: "Middleware", name: str, api_version: str, adapter: APIVersionsAdapter, *,
                 passthrough_nonexistent_methods=False):
        """
        :param middleware: `Middleware` instance
        :param name: method name
        :param api_version: API version name used to convert parameters and return value
        :param adapter: `APIVersionsAdapter` instance
        """
        super().__init__(middleware, name)
        self.api_version = api_version
        self.adapter = adapter
        # FIXME: Remove this when legacy WS API is removed
        self.passthrough_nonexistent_methods = passthrough_nonexistent_methods

        methodobj = self.methodobj
        if crud_methodobj := real_crud_method(methodobj):
            methodobj = crud_methodobj
        if hasattr(methodobj, "new_style_accepts"):  # FIXME: Remove this check when all models become new style
            self.accepts_model = methodobj.new_style_accepts
            self.returns_model = methodobj.new_style_returns
        else:
            self.accepts_model = None
            self.returns_model = None

    async def call(self, app: "RpcWebSocketApp", params):
        if self.accepts_model:  # FIXME: Remove this check when all models become new style
            params = self._adapt_params(params)

        return await super().call(app, params)

    def _adapt_params(self, params):
        try:
            legacy_accepts_model = self.adapter.versions[self.api_version].models[self.accepts_model.__name__]
        except KeyError:
            if self.passthrough_nonexistent_methods:
                return params

            # The legacy API does not contain signature definition for this method, which means it didn't exist
            # when that API was released.
            raise MethodNotFoundError(*self.name.rsplit(".", 1))

        params_dict = model_dict_from_list(legacy_accepts_model, params)

        adapted_params_dict = self.adapter.adapt(
            params_dict,
            legacy_accepts_model.__name__,
            self.api_version,
            self.adapter.current_version,
        )

        return [adapted_params_dict[field] for field in self.accepts_model.model_fields]

    def _dump_result(self, app: "RpcWebSocketApp", methodobj, result):
        if self.accepts_model:  # FIXME: Remove this check when all models become new style
            try:
                model, result = self.adapter.adapt_model(
                    {"result": result},
                    self.returns_model.__name__,
                    self.adapter.current_version,
                    self.api_version,
                )
            except APIVersionDoesNotContainModelException:
                if self.passthrough_nonexistent_methods:
                    return super()._dump_result(app, methodobj, result)

                raise

            return self.middleware.dump_result(self.serviceobj, methodobj, app, result["result"],
                                               new_style_returns_model=model)

        return super()._dump_result(app, methodobj, result)

    def dump_args(self, params):
        if self.accepts_model:  # FIXME: Remove this check when all models become new style
            return dump_params(self.accepts_model, params, False)

        return super().dump_args(params)
