import errno
import hashlib
import itertools
import os
import time

import requests
import requests.exceptions

from middlewared.service import CallError, private, Service
from middlewared.utils.size import format_size

from .utils import DOWNLOAD_UPDATE_FILE, scale_update_server


class UpdateService(Service):
    @private
    def download_impl_scale(self, job, train, location, progress_proportion):
        job.set_progress(0, "Retrieving update manifest")

        train_check = self.middleware.call_sync("update.check_train", train)
        if train_check["status"] == "AVAILABLE":
            dst = os.path.join(location, DOWNLOAD_UPDATE_FILE)
            try:
                with open(dst, "rb") as f:
                    job.set_progress(0, "Verifying existing update")
                    checksum = hashlib.file_digest(f, "sha256").hexdigest()
            except FileNotFoundError:
                pass
            else:
                if checksum == train_check["checksum"]:
                    return True
                else:
                    self.logger.warning("Invalid update file checksum %r, re-downloading", checksum)
                    os.unlink(dst)

            st = os.statvfs(location)
            avail = st.f_bavail * st.f_frsize

            # make sure we have at least as the filesize plus 500MiB
            required_size = train_check["filesize"] + 500 * 1024**2
            if required_size > avail:
                raise CallError(
                    f"{location}: insufficient available space: {format_size(avail)}, "
                    f"required: {format_size(required_size)}", errno.ENOSPC
                )

            for i in itertools.count(1):
                with open(dst, "ab") as f:
                    download_start = time.monotonic()
                    progress = None
                    try:
                        start = os.path.getsize(dst)
                        with requests.get(
                            f"{scale_update_server()}/{train}/{train_check['filename']}",
                            stream=True,
                            timeout=30,
                            headers={"Range": f"bytes={start}-"}
                        ) as r:
                            r.raise_for_status()
                            total = start + int(r.headers["Content-Length"])
                            for chunk in r.iter_content(chunk_size=8 * 1024 * 1024):
                                progress = f.tell()

                                job.set_progress(
                                    progress / total * progress_proportion,
                                    f'Downloading update: {format_size(total)} at '
                                    f'{format_size(progress / (time.monotonic() - download_start))}/s'
                                )

                                f.write(chunk)
                            break
                    except Exception as e:
                        if i < 5 and progress and any(ee in str(e) for ee in ("ECONNRESET", "ETIMEDOUT")):
                            self.middleware.logger.warning("Recoverable update download error: %r", e)
                            time.sleep(2)
                            continue

                        raise

            size = os.path.getsize(dst)
            if size != total:
                os.unlink(dst)
                raise CallError(f'Downloaded update file mismatch ({size} != {total})')

            return True

        return False
